require_relative '../../app/helpers/application_helper'
require 'timing_logger'

class ArticleLoader
  include ApplicationHelper

  def load_data(repo_path)
    @logger = TimingLogger.new(STDOUT, 'data:load:articles')
    Dir["#{repo_path}/articles/*.md"].each do |file|
      begin
        content = File.read(file)
        md = YAML.load(content)
        create_article(file, md, content) unless md['draft'] || bad_md?(file, md)
      rescue => e
        @logger.error("Exception in #{file}: #{e}")
        @logger.error(e.full_message)
      end
    end
    @logger.info('Done.')
  end

  def bad_md?(file, md)
    bad_md = md['title'].nil? or md['art'].nil? or md['blurb'].nil?
    if bad_md
      @logger.warn "INVALID ARTICLE: title, art, or blurb missing in #{file}"
    end
    return bad_md
  end

  def create_article(file, md, content)
    slug = file.match(%r{^.*/(?<slug>.+).md$}).named_captures['slug']
    article = Article.create!(
      title: md['title'],
      author: md['author'],
      art: md['art'],
      blurb: md['blurb'],
      fulltext: content.split("---")[2],
      slug: slug,
    )

    md['cves']&.each do |cve|
      vuln = Vulnerability.find_by(cve: cve)
      if vuln.nil?
        @logger.warn("BAD ARTICLE METADATA: can't find CVE #{cve} in #{file}")
      else
        ArticleVulnerability.create!(article: article, vulnerability: vuln)
      end
    end

    md['tags']&.each do |tag|
      db_tag = Tag.find_by(shortname: tag)
      if db_tag.nil?
        @logger.warn("BAD ARTICLE METADATA: can't find tag #{tag} in #{file}")
      else
        ArticleTag.create!(article: article, tag: db_tag)
      end
    end

    md['filepaths']&.each do |f|
      db_filepath = Filepath.find_by(filepath: f)
      if db_filepath.nil?
        @logger.warn("BAD ARTICLE METADATA: can't find filepath #{f} in #{file}")
      else
        ArticleFilepath.create!(article: article, filepath: db_filepath)
      end
    end
  end
end
